package main

import (
	"context"
	"fmt"
	"log"
	"net"
	"net/http"
	"os"

	"git.3crabs.ru/save_my_money/smm_core/internal/app"
	"git.3crabs.ru/save_my_money/smm_core/internal/services/budget"
	"git.3crabs.ru/save_my_money/smm_core/internal/services/category"
	"git.3crabs.ru/save_my_money/smm_core/internal/services/user"
	"github.com/grpc-ecosystem/grpc-gateway/v2/runtime"
	"github.com/jackc/pgx/v5/pgxpool"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/metadata"
	"google.golang.org/grpc/status"

	proto "git.3crabs.ru/save_my_money/smm_core/proto"
)

func main() {
	// Create a listener on TCP port
	lis, err := net.Listen("tcp", ":8080")
	if err != nil {
		log.Fatalln("Failed to listen:", err)
	}

	dbpool, err := pgxpool.New(context.Background(), os.Getenv("DATABASE_URL"))
	if err != nil {
		fmt.Fprintf(os.Stderr, "Unable to create connection pool: %v\n", err)
		os.Exit(1)
	}
	defer dbpool.Close()

	categoryService := category.NewCategoryService(dbpool)
	userService := user.NewUserService(dbpool)
	budgetService := budget.NewBudgetService(dbpool)

	// Create a gRPC server object
	s := grpc.NewServer(
		grpc.UnaryInterceptor(
			func(
				ctx context.Context,
				req any,
				info *grpc.UnaryServerInfo,
				handler grpc.UnaryHandler,
			) (resp any, err error) {
				md, ok := metadata.FromIncomingContext(ctx)
				if !ok || len(md["authorization"]) == 0 {
					return nil, status.Errorf(codes.Unauthenticated, "no auth token")
				}
				authToken := md["authorization"][0]
				if authToken != os.Getenv("AUTHORIZATION") {
					return nil, status.Errorf(codes.Unauthenticated, "invalid token")
				}
				return handler(ctx, req)
			},
		),
	)
	// Attach the Greeter service to the server
	proto.RegisterSmmCoreServer(
		s,
		app.NewServer(
			categoryService,
			userService,
			budgetService,
		),
	)
	// Serve gRPC server
	log.Println("Serving gRPC on 0.0.0.0:8080")
	go func() {
		log.Fatalln(s.Serve(lis))
	}()

	// Create a client connection to the gRPC server we just started
	// This is where the gRPC-Gateway proxies the requests
	conn, err := grpc.NewClient(
		"0.0.0.0:8080",
		grpc.WithTransportCredentials(insecure.NewCredentials()),
	)
	if err != nil {
		log.Fatalln("Failed to dial server:", err)
	}

	mux := runtime.NewServeMux(
		runtime.WithMetadata(func(ctx context.Context, request *http.Request) metadata.MD {
			header := request.Header.Get("User-Id")
			return metadata.Pairs("user-id", header)
		}),
	)
	// Register Greeter
	err = proto.RegisterSmmCoreHandler(context.Background(), mux, conn)
	if err != nil {
		log.Fatalln("Failed to register gateway:", err)
	}

	err = mux.HandlePath("GET", "/swagger", func(w http.ResponseWriter, r *http.Request, pathParams map[string]string) {
		data, err := os.ReadFile("resources/smm_core.swagger.json")
		if err != nil {
			w.WriteHeader(http.StatusInternalServerError)
		}
		_, _ = w.Write(data)
	})
	if err != nil {
		panic(err)
	}
	log.Println("Serving swagger json on http://0.0.0.0:8090/swagger")

	gwServer := &http.Server{
		Addr:    ":8090",
		Handler: cors(mux),
	}

	log.Println("Serving gRPC-Gateway on http://0.0.0.0:8090")
	log.Fatalln(gwServer.ListenAndServe())
}

func cors(h http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.Header().Set("Access-Control-Allow-Origin", "*")
		w.Header().Set("Access-Control-Allow-Methods", "GET, POST, PATCH, DELETE")
		w.Header().Set("Access-Control-Allow-Headers", "Accept, Content-Type, Content-Length, Accept-Encoding, Authorization, ResponseType")
		if r.Method == "OPTIONS" {
			return
		}
		h.ServeHTTP(w, r)
	})
}
